import copy
import json
from contextlib import contextmanager
from typing import (
    Deque,
    List,
    AsyncGenerator,
    Optional,
    Any,
    Dict,
    Iterator,
    override,
    Iterable,
    Generator,
)

from compressedfhir.utilities.compressed_dict.v1.compressed_dict import (
    CompressedDict,
)
from compressedfhir.utilities.fhir_json_encoder import FhirJSONEncoder


class BaseResourceList[T: CompressedDict[str, Any]](Deque[T]):
    """
    Represents a list of FHIR resources.
    """

    __slots__: List[str] = []

    def __init__(
        self,
        iterable: Iterable[T] | None = None,
        maxlen: Optional[int] = None,
    ) -> None:
        # if iterable is not None:
        # assert all([isinstance(r, CompressedDict) for r in iterable]), f"type {type(next(iter(iterable)))}"
        super().__init__(iterable, maxlen=maxlen) if iterable else super().__init__()

    @override
    def append(self, x: T, /) -> None:
        super().append(x)
        # assert isinstance(x, CompressedDict), f"{type(x)} is not a CompressedDict"

    def json(self) -> str:
        """
        Convert the list of FhirResource objects to a JSON string.

        :return: JSON string representation of the list.
        """
        return json.dumps([r.dict() for r in self], cls=FhirJSONEncoder)

    def get_resources(self) -> "BaseResourceList[T]":
        """
        Get all resources in the list.

        :return: A list of FhirResource objects.
        """
        # this is here to keep compatibility with FhirResourceMap
        return self

    async def consume_resource_batch_async(
        self,
        *,
        batch_size: Optional[int],
    ) -> AsyncGenerator["BaseResourceList[T]", None]:
        """
        Consume resources in batches asynchronously.

        :param batch_size: The size of each batch.
        :return: An async generator yielding batches of FhirResourceList.
        """
        batch: BaseResourceList[T]
        if batch_size is None:
            batch = BaseResourceList[T]()
            while self:
                batch.append(self.popleft())
            yield batch
        elif batch_size <= 0:
            raise ValueError("Batch size must be greater than 0.")
        else:
            while self:
                batch = BaseResourceList[T]()
                for _ in range(min(batch_size, len(self))):
                    batch.append(self.popleft())
                yield batch

    def consume_resource_batch(
        self,
        *,
        batch_size: Optional[int],
    ) -> Generator["BaseResourceList[T]", None, None]:
        """
        Consume resources in batches asynchronously.

        :param batch_size: The size of each batch.
        :return: An async generator yielding batches of FhirResourceList.
        """
        batch: BaseResourceList[T]
        if batch_size is None:
            batch = BaseResourceList[T]()
            while self:
                batch.append(self.popleft())
            yield batch
        elif batch_size <= 0:
            raise ValueError("Batch size must be greater than 0.")
        else:
            while self:
                batch = BaseResourceList[T]()
                for _ in range(min(batch_size, len(self))):
                    batch.append(self.popleft())
                yield batch

    def __deepcopy__(self, memo: Dict[int, Any]) -> "BaseResourceList[T]":
        """
        Create a copy of the FhirResourceList.

        :return: A new FhirResourceList object with the same resources.
        """
        return BaseResourceList[T]([copy.deepcopy(r) for r in self])

    def __repr__(self) -> str:
        """
        Return a string representation of the FhirResourceList.

        :return: A string representation of the FhirResourceList.
        """
        return f"FhirResourceList(resources: {len(self)})"

    @contextmanager
    def transaction(self) -> Iterator["BaseResourceList[T]"]:
        """
        Opens a transaction for all resources in the list.

        Deserializes the dictionary before entering the context
        Serializes the dictionary after exiting the context

        Raises:
            CompressedDictAccessError: If methods are called outside the context

        """
        try:
            self.start_transaction()

            yield self

        finally:
            self.end_transaction()

    def start_transaction(self) -> "BaseResourceList[T]":
        """
        Starts a transaction for each resource in the list.  Use transaction() for a contextmanager for simpler usage.
        """

        for resource in self:
            resource.start_transaction()
        return self

    def end_transaction(self) -> "BaseResourceList[T]":
        """
        Ends a transaction.  Use transaction() for a context_manager for simpler usage.

        """
        for resource in self:
            resource.end_transaction()
        return self



import json

from compressedfhir.fhir.fhir_bundle import FhirBundle
from compressedfhir.fhir.fhir_bundle_entry import FhirBundleEntry
from compressedfhir.fhir.fhir_bundle_entry_list import FhirBundleEntryList
from compressedfhir.fhir.fhir_resource import FhirResource
from compressedfhir.utilities.compressed_dict.v1.compressed_dict_storage_mode import (
    CompressedDictStorageMode,
)


class TestBundle:
    def test_init_minimal(self) -> None:
        """Test initialization with minimal parameters."""
        bundle = FhirBundle(type_="searchset")
        assert bundle.type_ == "searchset"
        assert len(bundle.entry) == 0
        assert bundle.total is None
        assert bundle.id_ is None
        assert bundle.timestamp is None

    def test_init_full(self) -> None:
        """Test initialization with all parameters."""
        entries = FhirBundleEntryList(
            [
                FhirBundleEntry(
                    resource={"resourceType": "Patient"},
                    request=None,
                    response=None,
                    storage_mode=CompressedDictStorageMode(),
                ),
                FhirBundleEntry(
                    resource={"resourceType": "Observation"},
                    request=None,
                    response=None,
                    storage_mode=CompressedDictStorageMode(),
                ),
            ]
        )
        bundle = FhirBundle(
            type_="searchset",
            entry=entries,
            total=2,
            id_="test-bundle",
            timestamp="2023-12-01T12:00:00Z",
        )
        assert bundle.type_ == "searchset"
        assert bundle.entry is not None
        assert len(bundle.entry) == 2
        assert bundle.total == 2
        assert bundle.id_ == "test-bundle"
        assert bundle.timestamp == "2023-12-01T12:00:00Z"

    def test_to_dict_minimal(self) -> None:
        """Test converting to dictionary with minimal parameters."""
        bundle = FhirBundle(type_="searchset")
        result = bundle.dict()
        assert result == {"type": "searchset", "resourceType": "Bundle"}

    def test_to_dict_full(self) -> None:
        """Test converting to dictionary with all parameters."""
        entries = FhirBundleEntryList(
            [
                FhirBundleEntry(
                    resource={"resourceType": "Patient"},
                    request=None,
                    response=None,
                    storage_mode=CompressedDictStorageMode(),
                ),
                FhirBundleEntry(
                    resource={"resourceType": "Observation"},
                    request=None,
                    response=None,
                    storage_mode=CompressedDictStorageMode(),
                ),
            ]
        )
        bundle = FhirBundle(
            type_="searchset",
            entry=entries,
            total=2,
            id_="test-bundle",
            timestamp="2023-12-01T12:00:00Z",
        )
        result = bundle.dict()
        assert result == {
            "type": "searchset",
            "resourceType": "Bundle",
            "id": "test-bundle",
            "timestamp": "2023-12-01T12:00:00Z",
            "total": 2,
            "entry": [
                {"resource": {"resourceType": "Patient"}},
                {"resource": {"resourceType": "Observation"}},
            ],
        }

    def test_to_json(self) -> None:
        """Test converting Bundle to JSON."""
        entries = FhirBundleEntryList(
            [
                FhirBundleEntry(
                    resource={"resourceType": "Patient"},
                    request=None,
                    response=None,
                    storage_mode=CompressedDictStorageMode(),
                ),
                FhirBundleEntry(
                    resource={"resourceType": "Observation"},
                    request=None,
                    response=None,
                    storage_mode=CompressedDictStorageMode(),
                ),
            ]
        )
        bundle = FhirBundle(
            type_="searchset",
            entry=entries,
            total=2,
            id_="test-bundle",
            timestamp="2023-12-01T12:00:00Z",
        )
        json_str = bundle.json()
        parsed_json = json.loads(json_str)
        assert parsed_json["type"] == "searchset"
        assert parsed_json["resourceType"] == "Bundle"

    def test_add_diagnostics_to_operation_outcomes(self) -> None:
        """Test adding diagnostics to OperationOutcome resources."""
        resource = FhirResource(
            initial_dict={
                "resourceType": "OperationOutcome",
                "issue": [{"severity": "error", "details": {}}],
            },
            storage_mode=CompressedDictStorageMode(),
        )
        diagnostics_coding = [{"code": "test-code"}]

        updated_resource = FhirBundle.add_diagnostics_to_operation_outcomes(
            resource=resource, diagnostics_coding=diagnostics_coding
        )

        with updated_resource.transaction():
            assert (
                updated_resource["issue"][0]["details"]["coding"] == diagnostics_coding
            )


class TestFhirBundleCopy:
    def test_copy_full_bundle(self) -> None:
        """
        Test copying a fully populated FhirBundle
        """
        # Create a mock FhirBundleEntryList
        mock_entry_list = FhirBundleEntryList()
        mock_resource = FhirBundleEntry(
            resource=FhirResource({"resourceType": "Patient", "id": "123"})
        )
        mock_entry_list.append(mock_resource)

        # Create original bundle
        original_bundle = FhirBundle(
            id_="test-bundle-id",
            timestamp="2023-01-01T00:00:00Z",
            type_="transaction",
            entry=mock_entry_list,
            total=1,
        )

        # Create a copy
        copied_bundle = original_bundle.copy()

        # Assert that the copied bundle has the same attributes
        assert copied_bundle.id_ == original_bundle.id_
        assert copied_bundle.timestamp == original_bundle.timestamp
        assert copied_bundle.type_ == original_bundle.type_
        assert copied_bundle.total == original_bundle.total

        # Ensure the entry list is a copy, not the same object
        assert copied_bundle.entry is not original_bundle.entry
        assert copied_bundle.entry is not None
        assert original_bundle.entry is not None
        assert len(copied_bundle.entry) == len(original_bundle.entry)
        assert copied_bundle.entry[0].dict() == original_bundle.entry[0].dict()

    def test_copy_empty_bundle(self) -> None:
        """
        Test copying a bundle with no entries
        """
        # Create an empty bundle
        original_bundle = FhirBundle(type_="batch", entry=None, total=None)

        # Create a copy
        copied_bundle = original_bundle.copy()

        # Assert that the copied bundle has the same attributes
        assert copied_bundle.id_ is None
        assert copied_bundle.timestamp is None
        assert copied_bundle.type_ == "batch"
        assert copied_bundle.total is None
        assert len(copied_bundle.entry) == 0

    def test_copy_modifying_original_does_not_affect_copy(self) -> None:
        """
        Test that modifying the original bundle does not affect the copy
        """
        # Create a mock FhirBundleEntryList
        mock_entry_list = FhirBundleEntryList()
        mock_resource = FhirResource({"resourceType": "Patient", "id": "123"})
        mock_entry_list.append(FhirBundleEntry(resource=mock_resource))

        # Create original bundle
        original_bundle = FhirBundle(
            id_="test-bundle-id",
            timestamp="2023-01-01T00:00:00Z",
            type_="transaction",
            entry=mock_entry_list,
            total=1,
        )

        # Create a copy
        copied_bundle = original_bundle.copy()

        # Modify the original bundle
        original_bundle.id_ = "modified-id"
        original_bundle.timestamp = "2023-02-01T00:00:00Z"
        original_bundle.total = 2

        # Assert that the copied bundle remains unchanged
        assert copied_bundle.id_ == "test-bundle-id"
        assert copied_bundle.timestamp == "2023-01-01T00:00:00Z"
        assert copied_bundle.total == 1

    def test_copy_returns_new_instance(self) -> None:
        """
        Test that copy() returns a new FhirBundle instance
        """
        # Create a bundle
        original_bundle = FhirBundle(type_="batch")

        # Create a copy
        copied_bundle = original_bundle.copy()

        # Assert that it's a different object
        assert copied_bundle is not original_bundle



import json
from typing import Any, Dict, List, cast, Union, Optional, OrderedDict, overload
from datetime import datetime, date

import orjson


class FhirClientJsonHelpers:
    def __init__(self) -> None:
        pass

    @staticmethod
    def json_serial(obj: Any) -> str:
        """JSON serializer for objects not serializable by default json code"""

        # https://stackoverflow.com/questions/11875770/how-to-overcome-datetime-datetime-not-json-serializable
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()
        return str(obj)

    @staticmethod
    def remove_empty_elements(
        d: List[Dict[str, Any]] | Dict[str, Any],
    ) -> List[Dict[str, Any]] | Dict[str, Any]:
        """
        Recursively remove empty lists, empty dicts, or None elements from a dictionary
        or a list of dictionaries
        :param d: dictionary or list of dictionaries
        :return: dictionary or list of dictionaries
        """

        def empty(x: Any) -> bool:
            # Check if the input is None, an empty list, an empty dict, or an empty string
            return (
                x is None
                or x == []
                or x == {}
                or (isinstance(x, str) and x.strip() == "")
            )

        if not isinstance(d, (dict, list)):
            return d
        elif isinstance(d, list):
            return [
                cast(Dict[str, Any], v)
                for v in (FhirClientJsonHelpers.remove_empty_elements(v) for v in d)
                if not empty(v)
            ]
        else:
            return {
                k: v
                for k, v in (
                    (k, FhirClientJsonHelpers.remove_empty_elements(v))
                    for k, v in d.items()
                )
                if not empty(v)
            }

    @staticmethod
    @overload
    def remove_empty_elements_from_ordered_dict(
        d: List[OrderedDict[str, Any]],
    ) -> List[OrderedDict[str, Any]]: ...

    @staticmethod
    @overload
    def remove_empty_elements_from_ordered_dict(
        d: OrderedDict[str, Any],
    ) -> OrderedDict[str, Any]: ...

    @staticmethod
    def remove_empty_elements_from_ordered_dict(
        d: List[OrderedDict[str, Any]] | OrderedDict[str, Any],
    ) -> List[OrderedDict[str, Any]] | OrderedDict[str, Any]:
        """
        Recursively remove empty lists, empty dicts, or None elements from a dictionary
        or a list of dictionaries
        :param d: dictionary or list of dictionaries
        :return: dictionary or list of dictionaries
        """

        def empty(x: Any) -> bool:
            # Check if the input is None, an empty list, an empty dict, or an empty string
            return (
                x is None
                or x == []
                or x == {}
                or (isinstance(x, str) and x.strip() == "")
            )

        if not isinstance(d, (OrderedDict, list, Dict)):
            return d
        elif isinstance(d, list):
            return [
                v
                for v in (
                    FhirClientJsonHelpers.remove_empty_elements_from_ordered_dict(v)
                    for v in d
                )
                if not empty(v)
            ]
        else:
            return OrderedDict[str, Any](
                {
                    k: v
                    for k, v in (
                        (
                            k,
                            FhirClientJsonHelpers.remove_empty_elements_from_ordered_dict(
                                v
                            ),
                        )
                        for k, v in d.items()
                    )
                    if not empty(v)
                }
            )

    @staticmethod
    def convert_dict_to_fhir_json(dict_: Dict[str, Any]) -> str:
        """
        Returns dictionary as json string


        :return:
        """
        instance_variables: Dict[str, Any] = cast(
            Dict[str, Any], FhirClientJsonHelpers.remove_empty_elements(dict_)
        )

        instance_variables_text: str = json.dumps(
            instance_variables, default=FhirClientJsonHelpers.json_serial
        )
        return instance_variables_text

    @staticmethod
    def orjson_dumps(
        obj: Any, indent: Optional[int] = None, sort_keys: bool = False
    ) -> str:
        """
        Wrapper for orjson.dumps() to mimic json.dumps() behavior

        Args:
            obj: Object to serialize
            indent: Optional indentation (note: orjson has limited indent support)
            sort_keys: Whether to sort dictionary keys

        Returns:
            JSON string
        """
        # Serialization options
        options = 0

        # Handle sorting keys
        if sort_keys:
            options |= orjson.OPT_SORT_KEYS

        # Handle indentation (limited support)
        if indent is not None:
            options |= orjson.OPT_INDENT_2  # Fixed indentation

        # Serialize to bytes
        json_bytes = orjson.dumps(obj, option=options)

        # Convert bytes to string
        return json_bytes.decode("utf-8")

    @staticmethod
    def orjson_loads(json_input: Union[str, bytes]) -> Any:
        """
        Safely load JSON with type flexibility
        """
        try:
            # Converts input to appropriate type if needed
            if isinstance(json_input, str):
                json_input = json_input.encode("utf-8")

            return orjson.loads(json_input)
        except (orjson.JSONDecodeError, TypeError) as e:
            print(f"JSON Parsing Error: {e}")
            return None



import pytest
import json

from compressedfhir.fhir.fhir_resource import FhirResource
from compressedfhir.fhir.fhir_resource_list import FhirResourceList
from compressedfhir.fhir.fhir_resource_map import FhirResourceMap
from compressedfhir.utilities.compressed_dict.v1.compressed_dict_storage_mode import (
    CompressedDictStorageMode,
)


class TestFhirResourceMap:
    def test_init_empty(self) -> None:
        """Test initialization with no initial dictionary."""
        resource_map = FhirResourceMap()
        assert resource_map._resource_map == {}

    def test_init_with_initial_dict(self) -> None:
        """Test initialization with an initial dictionary."""
        initial_dict = {
            "Patient": FhirResourceList(
                [
                    FhirResource(
                        {"id": "123", "resourceType": "Patient"},
                        storage_mode=CompressedDictStorageMode(),
                    )
                ]
            ),
            "Observation": FhirResourceList(
                [
                    FhirResource(
                        {"id": "456", "resourceType": "Observation"},
                        storage_mode=CompressedDictStorageMode(),
                    )
                ]
            ),
        }
        resource_map = FhirResourceMap(initial_dict=initial_dict)
        assert resource_map._resource_map == initial_dict

    def test_to_dict(self) -> None:
        """Test conversion of resource map to dictionary."""
        patient_resource = {"id": "123", "resourceType": "Patient"}
        initial_dict = {
            "Patient": FhirResourceList(
                [
                    FhirResource(
                        patient_resource, storage_mode=CompressedDictStorageMode()
                    )
                ]
            )
        }
        resource_map = FhirResourceMap(initial_dict=initial_dict)
        result = resource_map.dict()
        assert result == {"Patient": [patient_resource]}

    def test_get_existing_resource_type(self) -> None:
        """Test getting resources for an existing resource type."""
        patient_resources = FhirResourceList(
            [
                FhirResource(
                    {"id": "123", "resourceType": "Patient"},
                    storage_mode=CompressedDictStorageMode(),
                )
            ]
        )
        initial_dict = {"Patient": patient_resources}
        resource_map = FhirResourceMap(initial_dict=initial_dict)

        result = resource_map.get(resource_type="Patient")
        assert result == patient_resources

    def test_get_nonexistent_resource_type(self) -> None:
        """Test getting resources for a non-existent resource type."""
        resource_map = FhirResourceMap()
        result = resource_map.get(resource_type="Patient")
        assert result is None

    def test_setitem(self) -> None:
        """Test setting an item in the resource map."""
        resource_map = FhirResourceMap()
        patient_resources = FhirResourceList(
            [
                FhirResource(
                    {"id": "123", "resourceType": "Patient"},
                    storage_mode=CompressedDictStorageMode(),
                )
            ]
        )
        resource_map["Patient"] = patient_resources

        assert resource_map._resource_map["Patient"] == patient_resources

    def test_getitem(self) -> None:
        """Test getting an item from the resource map."""
        patient_resources = FhirResourceList(
            [
                FhirResource(
                    {"id": "123", "resourceType": "Patient"},
                    storage_mode=CompressedDictStorageMode(),
                )
            ]
        )
        resource_map = FhirResourceMap(initial_dict={"Patient": patient_resources})

        result = resource_map["Patient"]
        assert result == patient_resources

    def test_delitem(self) -> None:
        """Test deleting an item from the resource map."""
        patient_resources = FhirResourceList(
            [
                FhirResource(
                    {"id": "123", "resourceType": "Patient"},
                    storage_mode=CompressedDictStorageMode(),
                )
            ]
        )
        resource_map = FhirResourceMap(initial_dict={"Patient": patient_resources})

        del resource_map["Patient"]
        assert "Patient" not in resource_map._resource_map

    def test_delitem_nonexistent(self) -> None:
        """Test deleting a non-existent item raises KeyError."""
        resource_map = FhirResourceMap()
        with pytest.raises(KeyError):
            del resource_map["Patient"]

    def test_contains(self) -> None:
        """Test checking if a resource type exists in the map."""
        patient_resources = FhirResourceList(
            [
                FhirResource(
                    {"id": "123", "resourceType": "Patient"},
                    storage_mode=CompressedDictStorageMode(),
                )
            ]
        )
        resource_map = FhirResourceMap(initial_dict={"Patient": patient_resources})

        assert "Patient" in resource_map
        assert "Observation" not in resource_map

    def test_items(self) -> None:
        """Test getting all items from the resource map."""
        patient_resources = FhirResourceList(
            [
                FhirResource(
                    {"id": "123", "resourceType": "Patient"},
                    storage_mode=CompressedDictStorageMode(),
                )
            ]
        )
        obs_resources = FhirResourceList(
            [
                FhirResource(
                    {"id": "456", "resourceType": "Observation"},
                    storage_mode=CompressedDictStorageMode(),
                )
            ]
        )
        initial_dict = {"Patient": patient_resources, "Observation": obs_resources}
        resource_map = FhirResourceMap(initial_dict=initial_dict)

        items = resource_map.items()
        assert len(items) == 2
        assert ("Patient", patient_resources) in items
        assert ("Observation", obs_resources) in items

    def test_get_resource_count(self) -> None:
        """Test getting the total number of resources."""
        initial_dict = {
            "Patient": FhirResourceList(
                [
                    FhirResource({"id": "1"}, storage_mode=CompressedDictStorageMode()),
                    FhirResource({"id": "2"}, storage_mode=CompressedDictStorageMode()),
                ]
            ),
            "Observation": FhirResourceList(
                [FhirResource({"id": "3"}, storage_mode=CompressedDictStorageMode())]
            ),
        }
        resource_map = FhirResourceMap(initial_dict=initial_dict)

        assert resource_map.get_resource_count() == 3

    def test_clear(self) -> None:
        """Test clearing the resource map."""
        patient_resources = FhirResourceList(
            [
                FhirResource(
                    {"id": "123", "resourceType": "Patient"},
                    storage_mode=CompressedDictStorageMode(),
                )
            ]
        )
        resource_map = FhirResourceMap(initial_dict={"Patient": patient_resources})

        resource_map.clear()
        assert len(resource_map._resource_map) == 0

    def test_get_resource_type_and_ids(self) -> None:
        """Test getting resource type and IDs."""
        initial_dict = {
            "Patient": FhirResourceList(
                [
                    FhirResource(
                        {"id": "123"}, storage_mode=CompressedDictStorageMode()
                    ),
                    FhirResource(
                        {"id": "456"}, storage_mode=CompressedDictStorageMode()
                    ),
                ]
            ),
            "Observation": FhirResourceList(
                [FhirResource({"id": "789"}, storage_mode=CompressedDictStorageMode())]
            ),
        }
        resource_map = FhirResourceMap(initial_dict=initial_dict)

        result = resource_map.get_resource_type_and_ids()
        assert set(result) == {"Patient/123", "Patient/456", "Observation/789"}

    def test_get_operation_outcomes(self) -> None:
        """Test getting operation outcomes."""
        op_outcomes = FhirResourceList(
            [FhirResource({"id": "op1"}, storage_mode=CompressedDictStorageMode())]
        )
        initial_dict = {"OperationOutcome": op_outcomes}
        resource_map = FhirResourceMap(initial_dict=initial_dict)

        result = resource_map.get_operation_outcomes()
        assert result == op_outcomes

    def test_get_operation_outcomes_empty(self) -> None:
        """Test getting operation outcomes when none exist."""
        resource_map = FhirResourceMap()
        result = resource_map.get_operation_outcomes()
        assert result == FhirResourceList()

    def test_get_resources_except_operation_outcomes(self) -> None:
        """Test getting resources excluding operation outcomes."""
        initial_dict = {
            "Patient": FhirResourceList(
                [FhirResource({"id": "123"}, storage_mode=CompressedDictStorageMode())]
            ),
            "OperationOutcome": FhirResourceList(
                [FhirResource({"id": "op1"}, storage_mode=CompressedDictStorageMode())]
            ),
            "Observation": FhirResourceList(
                [FhirResource({"id": "456"}, storage_mode=CompressedDictStorageMode())]
            ),
        }
        resource_map = FhirResourceMap(initial_dict=initial_dict)

        result = resource_map.get_resources_except_operation_outcomes()
        assert len(result) == 2
        assert all(
            resource.get("resourceType") != "OperationOutcome" for resource in result
        )

    def test_to_json(self) -> None:
        """Test converting resource map to JSON."""
        initial_dict = {
            "Patient": FhirResourceList(
                [
                    FhirResource(
                        {"id": "123", "resourceType": "Patient"},
                        storage_mode=CompressedDictStorageMode(),
                    )
                ]
            )
        }
        resource_map = FhirResourceMap(initial_dict=initial_dict)

        json_str = resource_map.json()
        parsed_json = json.loads(json_str)
        assert parsed_json == {"Patient": [{"id": "123", "resourceType": "Patient"}]}

    def test_get_count_of_resource_types(self) -> None:
        """Test getting the count of unique resource types."""
        initial_dict = {
            "Patient": FhirResourceList(
                [FhirResource({"id": "123"}, storage_mode=CompressedDictStorageMode())]
            ),
            "Observation": FhirResourceList(
                [FhirResource({"id": "456"}, storage_mode=CompressedDictStorageMode())]
            ),
        }
        resource_map = FhirResourceMap(initial_dict=initial_dict)

        assert resource_map.get_count_of_resource_types() == 2



import json
from unittest.mock import Mock

import pytest

from compressedfhir.fhir.fhir_resource_list import FhirResourceList


class TestFhirResourceList:
    def test_get_resource_type_and_ids(self) -> None:
        # Create mock FhirResource objects
        mock_resources = [
            Mock(resource_type="Patient", id="123"),
            Mock(resource_type="Observation", id="456"),
        ]

        resource_list = FhirResourceList(mock_resources)

        # Check the returned resource type and ids
        expected_result = ["Patient/123", "Observation/456"]
        assert resource_list.get_resource_type_and_ids() == expected_result

    def test_get_operation_outcomes(self) -> None:
        # Create mock FhirResource objects with different resource types
        mock_resources = [
            Mock(resource_type="OperationOutcome", id="err1"),
            Mock(resource_type="Patient", id="123"),
            Mock(resource_type="OperationOutcome", id="err2"),
        ]

        resource_list = FhirResourceList(mock_resources)

        # Get operation outcomes
        operation_outcomes = resource_list.get_operation_outcomes()

        # Check the result
        assert len(operation_outcomes) == 2
        assert all(r.resource_type == "OperationOutcome" for r in operation_outcomes)

    def test_get_resources_except_operation_outcomes(self) -> None:
        # Create mock FhirResource objects with different resource types
        mock_resources = [
            Mock(resource_type="OperationOutcome", id="err1"),
            Mock(resource_type="Patient", id="123"),
            Mock(resource_type="Observation", id="456"),
            Mock(resource_type="OperationOutcome", id="err2"),
        ]

        resource_list = FhirResourceList(mock_resources)

        # Get resources except operation outcomes
        valid_resources = resource_list.get_resources_except_operation_outcomes()

        # Check the result
        assert len(valid_resources) == 2
        assert all(r.resource_type != "OperationOutcome" for r in valid_resources)

    def test_remove_duplicates(self) -> None:
        # Create mock FhirResource objects with duplicates
        mock_resources = [
            Mock(resource_type="Patient", id="123", resource_type_and_id="Patient/123"),
            Mock(resource_type="Patient", id="123", resource_type_and_id="Patient/123"),
            Mock(
                resource_type="Observation",
                id="456",
                resource_type_and_id="Observation/456",
            ),
            Mock(resource_type="Patient", id="789", resource_type_and_id="Patient/789"),
        ]

        resource_list = FhirResourceList(mock_resources)

        # Remove duplicates
        resource_list.remove_duplicates()

        # Check the result
        assert len(resource_list) == 3
        assert len(set(r.resource_type_and_id for r in resource_list)) == 3

    def test_to_json(self) -> None:
        # Create mock FhirResource objects
        mock_resources = [
            Mock(
                resource_type="Patient",
                id="123",
                dict=lambda: {"id": "123", "resourceType": "Patient"},
            ),
            Mock(
                resource_type="Observation",
                id="456",
                dict=lambda: {"id": "456", "resourceType": "Observation"},
            ),
        ]

        resource_list = FhirResourceList(mock_resources)

        # Convert to JSON
        json_str = resource_list.json()

        # Parse and check the JSON
        parsed_json = json.loads(json_str)
        assert len(parsed_json) == 2
        assert parsed_json[0]["resourceType"] == "Patient"
        assert parsed_json[1]["resourceType"] == "Observation"

    @pytest.mark.asyncio
    async def test_consume_resource_async_default(self) -> None:
        # Create mock FhirResource objects
        mock_resources = [
            Mock(resource_type="Patient", id="123"),
            Mock(resource_type="Observation", id="456"),
        ]

        resource_list = FhirResourceList(mock_resources)

        # Consume resources asynchronously with default (None) batch size
        async for batch in resource_list.consume_resource_batch_async(batch_size=None):
            assert len(batch) == 2

        # Ensure all resources are consumed
        assert len(resource_list) == 0

    @pytest.mark.asyncio
    async def test_consume_resource_async_with_batch_size(self) -> None:
        # Create mock FhirResource objects
        mock_resources = [
            Mock(resource_type="Patient", id="123"),
            Mock(resource_type="Observation", id="456"),
            Mock(resource_type="Condition", id="789"),
        ]

        resource_list = FhirResourceList(mock_resources)

        # Consume resources asynchronously with batch size of 2
        batches = []
        async for batch in resource_list.consume_resource_batch_async(batch_size=2):
            batches.append(batch)

        # Check batches
        assert len(batches) == 2
        assert len(batches[0]) == 2
        assert len(batches[1]) == 1

        # Ensure all resources are consumed
        assert len(resource_list) == 0

    def test_consume_resource_async_invalid_batch_size(self) -> None:
        resource_list = FhirResourceList()

        # Test invalid batch sizes
        with pytest.raises(ValueError, match="Batch size must be greater than 0."):

            async def test() -> None:
                async for _ in resource_list.consume_resource_batch_async(batch_size=0):
                    pass

            # Run the async function
            import asyncio

            asyncio.run(test())



import copy
import json
from collections.abc import KeysView, ValuesView, ItemsView, MutableMapping
from contextlib import contextmanager
from typing import Dict, Optional, Iterator, cast, List, Any, overload, OrderedDict

import msgpack
import zlib

from compressedfhir.utilities.compressed_dict.v1.compressed_dict_access_error import (
    CompressedDictAccessError,
)
from compressedfhir.utilities.compressed_dict.v1.compressed_dict_storage_mode import (
    CompressedDictStorageMode,
    CompressedDictStorageType,
)
from compressedfhir.utilities.fhir_json_encoder import FhirJSONEncoder


class CompressedDict[K, V](MutableMapping[K, V]):
    """
    A dictionary-like class that supports flexible storage options.

    It can store data in raw format, MessagePack format, or compressed MessagePack format.
    """

    # use slots to reduce memory usage
    __slots__ = [
        "_storage_mode",
        "_working_dict",
        "_raw_dict",
        "_serialized_dict",
        "_properties_to_cache",
        "_cached_properties",
        "_length",
        "_transaction_depth",
    ]

    def __init__(
        self,
        *,
        initial_dict: Dict[K, V] | OrderedDict[K, V] | None = None,
        storage_mode: CompressedDictStorageMode,
        properties_to_cache: List[K] | None,
    ) -> None:
        """
        Initialize a dictionary with flexible storage options
        Args:
            initial_dict: Initial dictionary to store
            storage_mode: Storage method for dictionary contents
                - 'raw': Store as original dictionary
                - 'msgpack': Store as MessagePack serialized bytes
                - 'compressed_msgpack': Store as compressed MessagePack bytes
        """
        # Storage configuration
        self._storage_mode: CompressedDictStorageMode = storage_mode

        # Working copy of the dictionary during context
        self._working_dict: Optional[OrderedDict[K, V]] = None

        # Private storage options
        self._raw_dict: OrderedDict[K, V] = OrderedDict[K, V]()
        self._serialized_dict: Optional[bytes] = None

        self._properties_to_cache: List[K] | None = properties_to_cache

        self._cached_properties: Dict[K, V] = {}

        self._length: int = 0

        self._transaction_depth: int = 0

        # Populate initial dictionary if provided
        if initial_dict:
            # Ensure we use an OrderedDict to maintain original order
            initial_dict_ordered = (
                initial_dict
                if isinstance(initial_dict, OrderedDict)
                else OrderedDict[K, V](initial_dict)
            )
            self.replace(value=initial_dict_ordered)

    @contextmanager
    def transaction(self) -> Iterator["CompressedDict[K, V]"]:
        """
        Context manager to safely access and modify the dictionary.

        Deserializes the dictionary before entering the context
        Serializes the dictionary after exiting the context

        Raises:
            CompressedDictAccessError: If methods are called outside the context
        """
        try:
            self.start_transaction()

            # Yield the working dictionary
            yield self

        finally:
            self.end_transaction()

    def start_transaction(self) -> "CompressedDict[K, V]":
        """
        Starts a transaction.  Use transaction() for a contextmanager for simpler usage.
        """
        # Increment transaction depth
        self._transaction_depth += 1
        # Ensure working dictionary is ready on first entry
        if self._transaction_depth == 1:
            self.ensure_working_dict()

        return self

    def end_transaction(self) -> "CompressedDict[K, V]":
        """
        Ends a transaction.  Use transaction() for a context_manager for simpler usage.

        """
        # Decrement transaction depth
        self._transaction_depth -= 1
        # Only update serialized dict when outermost transaction completes
        if self._transaction_depth == 0:
            self._update_serialized_dict(current_dict=self._working_dict)

            # Clear the working dictionary
            self._working_dict = None

        return self

    def ensure_working_dict(self) -> None:
        """
        Ensures that the working dictionary is initialized and deserialized.

        """
        if not self._working_dict:
            self._working_dict = self.create_working_dict()

    def create_working_dict(self) -> OrderedDict[K, V]:
        working_dict: OrderedDict[K, V]
        # Deserialize the dictionary before entering the context
        if self._storage_mode.storage_type == "raw":
            # For raw mode, create a deep copy of the existing dictionary
            working_dict = self._raw_dict
        else:
            # For serialized modes, deserialize
            working_dict = (
                self._deserialize_dict(
                    serialized_dict=self._serialized_dict,
                    storage_type=self._storage_mode.storage_type,
                )
                if self._serialized_dict
                else OrderedDict[K, V]()
            )
            assert isinstance(working_dict, OrderedDict)
        return working_dict

    @staticmethod
    def _serialize_dict(
        *, dictionary: OrderedDict[K, V], storage_type: CompressedDictStorageType
    ) -> bytes:
        """
        Serialize entire dictionary using MessagePack

        Args:
            dictionary: Dictionary to serialize
            storage_type: Storage type to use for serialization

        Returns:
            Serialized bytes
        """
        assert isinstance(dictionary, OrderedDict)
        if storage_type == "compressed":
            # Serialize to JSON and compress with zlib
            json_str = json.dumps(
                dictionary, separators=(",", ":"), cls=FhirJSONEncoder
            )  # Most compact JSON representation
            return zlib.compress(
                json_str.encode("utf-8"), level=zlib.Z_BEST_COMPRESSION
            )

        # Serialize using MessagePack
        packed = msgpack.packb(
            dictionary,
            use_bin_type=True,  # Preserve string/bytes distinction
            use_single_float=True,  # More compact float representation
        )

        # Optional compression
        if storage_type == "compressed_msgpack":
            packed = zlib.compress(packed, level=zlib.Z_BEST_COMPRESSION)

        return packed

    @staticmethod
    def _deserialize_dict(
        *,
        serialized_dict: bytes,
        storage_type: CompressedDictStorageType,
    ) -> OrderedDict[K, V]:
        """
        Deserialize entire dictionary from MessagePack

        Args:
            serialized_dict: Serialized dictionary bytes

        Returns:
            Deserialized dictionary
        """
        assert serialized_dict is not None, "Serialized dictionary cannot be None"

        if storage_type == "compressed":
            # Decompress and parse JSON
            decompressed = zlib.decompress(serialized_dict)
            decoded_text = decompressed.decode("utf-8")
            # noinspection PyTypeChecker
            decompressed_dict = json.loads(decoded_text, object_pairs_hook=OrderedDict)
            assert isinstance(decompressed_dict, OrderedDict)
            return cast(OrderedDict[K, V], decompressed_dict)

        # Decompress if needed
        to_unpack = (
            zlib.decompress(serialized_dict)
            if storage_type == "compressed_msgpack"
            else serialized_dict
        )

        # Deserialize
        unpacked_dict = msgpack.unpackb(
            to_unpack,
            raw=False,  # Convert to strings
            strict_map_key=False,  # Handle potential key type variations
        )
        unpacked_dict = (
            unpacked_dict
            if isinstance(unpacked_dict, OrderedDict)
            else OrderedDict[K, V](unpacked_dict)
        )
        assert isinstance(unpacked_dict, OrderedDict)
        return cast(
            OrderedDict[K, V],
            unpacked_dict,
        )

    def _get_dict(self) -> OrderedDict[K, V]:
        """
        Get the dictionary, deserializing if necessary

        Returns:
            Current dictionary state
        """

        if self._working_dict is None:
            raise CompressedDictAccessError(
                "Dictionary access is only allowed within an transaction() block. "
                "Use 'with compressed_dict.transaction() as d:' to access the dictionary."
                f"You tried to access it with storage type {self._storage_mode.storage_type}."
            )

        if self._storage_mode.storage_type == "raw":
            return self._raw_dict

        # For non-raw modes, do not keep deserialized dict
        return self._working_dict

    def __getitem__(self, key: K) -> V:
        """
        Retrieve a value

        Args:
            key: Dictionary key

        Returns:
            Value associated with the key
        """

        if self._properties_to_cache and key in self._properties_to_cache:
            return self._cached_properties[key]

        if self._working_dict is None:
            raise CompressedDictAccessError(
                "Dictionary access is only allowed within an transaction() block. "
                "Use 'with compressed_dict.transaction() as d:' to access the dictionary."
            )
        return self._get_dict()[key]

    def __setitem__(self, key: K, value: V) -> None:
        """
        Set a value

        Args:
            key: Dictionary key
            value: Value to store
        """
        if self._working_dict is None:
            raise CompressedDictAccessError(
                "Dictionary modification is only allowed within an transaction() block. "
                "Use 'with compressed_dict.transaction() as d:' to modify the dictionary."
            )

        # Update the working dictionary
        self._working_dict[key] = value

        if self._properties_to_cache and key in self._properties_to_cache:
            # Update the cached properties if the key is in the list
            self._cached_properties[key] = value

    def _update_serialized_dict(self, current_dict: OrderedDict[K, V] | None) -> None:
        if current_dict is None:
            self._cached_properties.clear()
            self._length = 0
            self._serialized_dict = None
            self._raw_dict = OrderedDict[K, V]()
            return

        if self._properties_to_cache:
            for key in self._properties_to_cache:
                if key in current_dict:
                    self._cached_properties[key] = current_dict[key]

        self._length = len(current_dict)

        if self._working_dict is not None:
            # If the working dictionary is None, initialize it
            self._working_dict = current_dict

        if self._transaction_depth == 0:
            # If we're in a transaction,
            # The serialized dictionary will be updated after the transaction
            if self._storage_mode.storage_type == "raw":
                self._raw_dict = current_dict
            else:
                self._serialized_dict = (
                    self._serialize_dict(
                        dictionary=current_dict,
                        storage_type=self._storage_mode.storage_type,
                    )
                    if current_dict
                    else None
                )

    def __delitem__(self, key: K) -> None:
        """
        Delete an item

        Args:
            key: Key to delete
        """
        if self._working_dict is None:
            raise CompressedDictAccessError(
                "Dictionary modification is only allowed within an transaction() block. "
                "Use 'with compressed_dict.transaction() as d:' to modify the dictionary."
            )

        del self._working_dict[key]

    def __contains__(self, key: object) -> bool:
        """
        Check if a key exists

        Args:
            key: Key to check

        Returns:
            Whether the key exists
        """
        # first check if the key is in the cached properties
        if self._properties_to_cache and key in self._properties_to_cache:
            return self._cached_properties.__contains__(key)

        return self._get_dict().__contains__(key)

    def __len__(self) -> int:
        """
        Get the number of items

        Returns:
            Number of items in the dictionary
        """
        return self._length

    def __iter__(self) -> Iterator[K]:
        """
        Iterate over keys

        Returns:
            Iterator of keys
        """
        if self._working_dict is None:
            raise CompressedDictAccessError(
                "Dictionary modification is only allowed within an transaction() block. "
                "Use 'with compressed_dict.transaction() as d:' to modify the dictionary."
            )
        return iter(self._get_dict())

    def keys(self) -> KeysView[K]:
        """
        Get an iterator of keys

        Returns:
            Iterator of keys
        """
        return self._get_dict().keys()

    def values(self) -> ValuesView[V]:
        """
        Get an iterator of values

        Returns:
            Iterator of values
        """
        return self._get_dict().values()

    def items(self) -> ItemsView[K, V]:
        """
        Get an iterator of key-value pairs

        Returns:
            Iterator of key-value pairs
        """
        return self._get_dict().items()

    def dict(self, *, remove_nulls: bool = True) -> OrderedDict[K, V]:
        """
        Convert to a standard dictionary

        Returns:
            Standard dictionary with all values
        """
        if self._working_dict:
            return self._working_dict
        else:
            # if the working dict is None, return it but don't store it in the self._working_dict to keep memory low
            return self.create_working_dict()

    def __repr__(self) -> str:
        """
        String representation of the dictionary

        Returns:
            String representation
        """
        cached_property_list: List[str] = [
            f"{k}={v}" for k, v in self._cached_properties.items()
        ]
        return (
            (
                f"CompressedDict(storage_type='{self._storage_mode.storage_type}', keys={self._length}"
            )
            + (", " if cached_property_list else "")
            + ", ".join(cached_property_list)
            + ")"
        )

    def replace(
        self, *, value: Dict[K, V] | OrderedDict[K, V]
    ) -> "CompressedDict[K, V]":
        """
        Replace the current dictionary with a new one

        Args:
            value: New dictionary to store

        Returns:
            Self
        """
        if not value:
            self.clear()
            return self

        new_dict: OrderedDict[K, V] = (
            value if isinstance(value, OrderedDict) else OrderedDict[K, V](value)
        )
        self._update_serialized_dict(current_dict=new_dict)
        return self

    def clear(self) -> None:
        """
        Clear the dictionary
        """
        self._update_serialized_dict(current_dict=None)

    def __eq__(self, other: object) -> bool:
        """
        Check equality with another dictionary

        Args:
            other: Dictionary to compare with (CompressedDict or plain dict)

        Returns:
            Whether the dictionaries are equal in keys and values
        """
        # If other is not a dictionary-like object, return False
        if not isinstance(other, (CompressedDict, dict, OrderedDict)):
            return False

        # Get the dictionary representation of self
        self_dict = self.dict()

        # If other is a CompressedDict, use its _get_dict() method
        if isinstance(other, CompressedDict):
            other_dict = other.dict()
        else:
            # If other is a plain dict, use it directly
            if isinstance(other, OrderedDict):
                # If other is an OrderedDict, use it directly
                other_dict = other
            else:
                other_dict = OrderedDict[K, V](other)

        # Compare keys and values
        # Check that all keys in both dictionaries match exactly
        return set(self_dict.keys()) == set(other_dict.keys()) and all(
            self_dict[key] == other_dict[key] for key in self_dict
        )

    @overload
    def get(self, key: K) -> Optional[V]:
        """
        Get a value for an existing key

        :param key: Key to retrieve
        :return: Value or None if key is not found
        """
        ...

    # noinspection PyMethodOverriding
    @overload
    def get[_T](self, key: K, /, default: V | _T) -> V | _T:
        """
        Get a value with a default

        Args:
            key: Key to retrieve
            default: Default value if key is not found

        Returns:
            Value associated with the key or default
        """
        ...

    def get[_T](self, key: K, default: V | _T | None = None) -> V | _T | None:
        if key in self:
            return self[key]
        return default

    def __deepcopy__(self, memo: Dict[int, Any]) -> "CompressedDict[K, V]":
        """
        Create a deep copy of the dictionary

        Args:
            memo: Memoization dictionary for deep copy

        Returns:
            Deep copy of the dictionary
        """
        # Create a new instance with the same storage mode
        new_instance = CompressedDict(
            initial_dict=copy.deepcopy(self.dict()),
            storage_mode=self._storage_mode,
            properties_to_cache=self._properties_to_cache,
        )
        return new_instance

    def copy(self) -> "CompressedDict[K,V]":
        """
        Creates a copy of the BundleEntry object.

        :return: A new BundleEntry object with the same attributes.
        """
        return copy.deepcopy(self)

    def get_storage_mode(self) -> CompressedDictStorageMode:
        """
        Get the storage mode

        Returns:
            Storage mode
        """
        return self._storage_mode

    @overload
    def pop(self, key: K, /) -> V:
        """
        Remove and return a value for an existing key

        :param key: Key to remove
        :return: Removed value
        :raises KeyError: If key is not found
        """
        ...

    # noinspection PyMethodOverriding
    @overload
    def pop[_T](self, key: K, /, default: _T) -> V | _T:
        """
        Remove and return a value, or return default if key is not found

        :param key: Key to remove
        :param default: Default value to return if key is not found
        :return: Removed value or default
        """
        ...

    def pop[_T](self, key: K, /, default: V | _T | None = None) -> V | _T | None:
        """
        Remove and return a value

        :param key: Key to remove
        :param default: Optional default value if key is not found
        :return: Removed value or default
        :raises KeyError: If key is not found and no default is provided
        """
        if self._working_dict is None:
            raise CompressedDictAccessError(
                "Dictionary modification is only allowed within a transaction() block. "
                "Use 'with compressed_dict.transaction() as d:' to modify the dictionary."
            )

        # If no default is provided, use the standard dict.pop() behavior
        if default is None:
            return self._working_dict.pop(key)

        return self._working_dict.pop(key, default)

    def to_plain_dict(self) -> Dict[K, V]:
        """
        Get the plain dictionary representation

        Returns:
            Plain dictionary
        """
        return cast(
            Dict[K, V], json.loads(json.dumps(self.dict(), cls=FhirJSONEncoder))
        )



import dataclasses
import json
from datetime import datetime, date
from decimal import Decimal
from enum import Enum
from typing import Any


class FhirJSONEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)  # type:ignore
        if isinstance(o, Enum):
            return o.value
        if isinstance(o, Decimal):
            # Custom Decimal conversion
            if o == o.to_integral_value():
                # If Decimal is a whole number, return as int
                return int(o)
            else:
                # If Decimal has a non-zero decimal part, return as float
                return float(o)
        if isinstance(o, bytes):
            return o.decode("utf-8")
        if isinstance(o, (datetime, date)):
            return o.isoformat()
        if hasattr(o, "to_dict"):
            return o.dict()

        return super().default(o)



from typing import (
    List,
    Set,
    Optional,
    AsyncGenerator,
    cast,
    Generator,
    Dict,
    override,
    Iterable,
)

from compressedfhir.fhir.base_resource_list import BaseResourceList
from compressedfhir.fhir.fhir_resource import FhirResource


class FhirResourceList(BaseResourceList[FhirResource]):
    """
    Represents a list of FHIR resources.
    """

    __slots__: List[str] = BaseResourceList.__slots__

    def get_resource_type_and_ids(self) -> List[str]:
        """
        Get the resource type and IDs of the resources in the list.

        :return: A list of strings representing the resource type and IDs.
        """
        resource_type_and_ids: List[str] = []
        for resource in self:
            resource_type_and_ids.append(f"{resource.resource_type}/{resource.id}")
        return resource_type_and_ids

    def get_operation_outcomes(self) -> "FhirResourceList":
        """
        Gets the operation outcomes from the response

        :return: list of operation outcomes
        """
        return FhirResourceList(
            [r for r in self if r.resource_type == "OperationOutcome"]
        )

    def get_resources_except_operation_outcomes(self) -> "FhirResourceList":
        """
        Gets the normal FHIR resources by skipping any OperationOutcome resources

        :return: list of valid resources
        """
        return FhirResourceList(
            [r for r in self if r.resource_type != "OperationOutcome"]
        )

    def remove_duplicates(self) -> None:
        """
        removes duplicate resources from the response i.e., resources with same resourceType and id

        """

        # remove duplicates from the list if they have the same resourceType and id
        seen: Set[str] = set()
        i = 0
        while i < len(self):
            # Create a tuple of values for specified keys
            comparison_key = self[i].resource_type_and_id

            if not comparison_key:
                # Skip if resourceType or id is None
                i += 1
                continue

            if comparison_key in seen or self[i].id is None:
                # Remove duplicate entry
                self.remove(self[i])
            else:
                seen.add(comparison_key)
                i += 1

    async def consume_resource_batch_async(
        self,
        *,
        batch_size: Optional[int],
    ) -> AsyncGenerator["FhirResourceList", None]:
        async for r in super().consume_resource_batch_async(batch_size=batch_size):
            yield cast(FhirResourceList, r)

    def consume_resource_batch(
        self,
        *,
        batch_size: Optional[int],
    ) -> Generator["FhirResourceList", None, None]:
        for r in super().consume_resource_batch(batch_size=batch_size):
            yield cast(FhirResourceList, r)

    def get_count_by_resource_type(self) -> Dict[str, int]:
        """
        Gets the count of resources by resource type.

        :return: The count of resources by resource type
        """
        resources_by_type: Dict[str, int] = dict()
        if len(self) == 0:
            return resources_by_type

        entry: FhirResource
        for resource in [r for r in self if r is not None]:
            resource_type: str = resource.resource_type or "unknown"
            if resource_type not in resources_by_type:
                resources_by_type[resource_type] = 0
            resources_by_type[resource_type] += 1
        return resources_by_type

    @override
    def append(self, x: FhirResource, /) -> None:
        """
        Append an entry to the FhirBundleEntryList.

        :param x: The entry to append.
        """
        if not isinstance(x, FhirResource):
            raise TypeError("Only FhirBundleEntry instances can be appended.")

        # check that we don't have a duplicate entry
        key: Optional[str] = x.resource_type_and_id
        if key is None:
            super().append(x)
        else:
            for entry in self:
                if entry.resource_type_and_id == key:
                    # we have a duplicate entry
                    return
            super().append(x)

    @override
    def extend(self, iterable: Iterable[FhirResource], /) -> None:
        """
        Extend the FhirBundleEntryList with entries from an iterable.

        :param iterable: The iterable containing entries to extend.
        """
        for entry in iterable:
            self.append(entry)



import copy
from typing import Deque, Optional, AsyncGenerator, List, Any, Dict, override, Iterable

from compressedfhir.fhir.fhir_bundle_entry import FhirBundleEntry


class FhirBundleEntryList(Deque[FhirBundleEntry]):
    """
    Represents a list of FHIR Bundle entries.
    """

    __slots__: List[str] = []

    async def consume_resource_async(
        self,
        *,
        batch_size: Optional[int],
    ) -> AsyncGenerator["FhirBundleEntryList", None]:
        """
        Consume bundle entries in batches asynchronously.

        :param batch_size: The size of each batch.
        :return: An async generator yielding batches of FhirResourceList.
        """
        if batch_size is None:
            while self:
                yield FhirBundleEntryList([self.popleft()])
        elif batch_size <= 0:
            raise ValueError("Batch size must be greater than 0.")
        else:
            while self:
                batch = FhirBundleEntryList()
                for _ in range(min(batch_size, len(self))):
                    batch.append(self.popleft())
                yield batch

    def __deepcopy__(self, memo: Dict[int, Any]) -> "FhirBundleEntryList":
        """
        Create a shallow copy of the FhirBundleEntryList.

        :return: A new FhirBundleEntryList instance with the same entries.
        """
        return FhirBundleEntryList([copy.deepcopy(entry) for entry in self])

    def __repr__(self) -> str:
        """
        Return a string representation of the FhirBundleEntryList.

        :return: A string representation of the FhirBundleEntryList.
        """
        return f"FhirBundleEntryList(entries: {len(self)})"

    @override
    def append(self, x: FhirBundleEntry, /) -> None:
        """
        Append an entry to the FhirBundleEntryList.

        :param x: The entry to append.
        """
        if not isinstance(x, FhirBundleEntry):
            raise TypeError("Only FhirBundleEntry instances can be appended.")

        # check that we don't have a duplicate entry
        key: Optional[str] = x.resource_type_and_id
        if key is None:
            super().append(x)
        else:
            for entry in self:
                if entry.resource_type_and_id == key:
                    # we have a duplicate entry
                    return
            super().append(x)

    @override
    def extend(self, iterable: Iterable[FhirBundleEntry], /) -> None:
        """
        Extend the FhirBundleEntryList with entries from an iterable.

        :param iterable: The iterable containing entries to extend.
        """
        for entry in iterable:
            self.append(entry)



import dataclasses
from typing import Literal, TypeAlias

CompressedDictStorageType: TypeAlias = Literal[
    "raw", "compressed", "msgpack", "compressed_msgpack"
]
"""
CompressedDictStorageType is a type alias for the different storage types
raw: No compression
compressed: Compressed using zlib
msgpack: Compressed using msgpack
compressed_msgpack: Compressed using msgpack with zlib
"""


@dataclasses.dataclass(slots=True)
class CompressedDictStorageMode:
    """
    CompressedDictStorageMode is a dataclass that defines the storage mode
    """

    storage_type: CompressedDictStorageType = "compressed"
    """
    CompressedDictStorageType is a type alias for the different storage types
    raw: No compression
    compressed: Compressed using zlib
    msgpack: Compressed using msgpack
    compressed_msgpack: Compressed using msgpack with zlib
    """

    @classmethod
    def default(cls) -> "CompressedDictStorageMode":
        """
        Returns the default storage mode
        """
        return cls(storage_type="compressed")

    @classmethod
    def raw(cls) -> "CompressedDictStorageMode":
        """
        Returns the default storage mode
        """
        return cls(storage_type="raw")

    @classmethod
    def compressed(cls) -> "CompressedDictStorageMode":
        """
        Returns the default storage mode
        """
        return cls(storage_type="compressed")



from datetime import datetime, timezone

from compressedfhir.fhir.fhir_bundle_entry import FhirBundleEntry
from compressedfhir.fhir.fhir_bundle_entry_request import FhirBundleEntryRequest
from compressedfhir.fhir.fhir_bundle_entry_response import (
    FhirBundleEntryResponse,
)
from compressedfhir.utilities.compressed_dict.v1.compressed_dict_storage_mode import (
    CompressedDictStorageMode,
)


class TestBundleEntry:
    def test_init_minimal(self) -> None:
        """Test initialization with minimal parameters."""
        entry = FhirBundleEntry(
            resource={"resourceType": "Patient"},
            request=None,
            response=None,
            storage_mode=CompressedDictStorageMode(),
        )
        assert entry.resource is not None
        with entry.resource.transaction():
            assert entry.resource.dict() == {"resourceType": "Patient"}
            assert entry.request is None
            assert entry.response is None
            assert entry.fullUrl is None

    def test_init_full(self) -> None:
        """Test initialization with all parameters."""
        resource = {"resourceType": "Patient", "id": "123"}
        request = FhirBundleEntryRequest(url="https://example.com")
        response = FhirBundleEntryResponse(status="200", etag=None, lastModified=None)

        entry = FhirBundleEntry(
            resource=resource,
            request=request,
            response=response,
            fullUrl="https://example.com/Patient/123",
            storage_mode=CompressedDictStorageMode(),
        )
        assert entry.resource is not None
        with entry.resource.transaction():
            assert entry.resource.dict() == resource
            assert entry.request == request
            assert entry.response == response
            assert entry.fullUrl == "https://example.com/Patient/123"

    def test_to_dict_minimal(self) -> None:
        """Test converting to dictionary with minimal parameters."""
        entry = FhirBundleEntry(
            resource={"resourceType": "Patient"},
            request=None,
            response=None,
            storage_mode=CompressedDictStorageMode(),
        )
        result = entry.dict()
        assert result == {"resource": {"resourceType": "Patient"}}

    def test_to_dict_full(self) -> None:
        """Test converting to dictionary with all parameters."""
        resource = {"resourceType": "Patient", "id": "123"}
        request = FhirBundleEntryRequest(url="https://example.com")
        response = FhirBundleEntryResponse(status="200", etag=None, lastModified=None)

        entry = FhirBundleEntry(
            resource=resource,
            request=request,
            response=response,
            fullUrl="https://example.com/Patient/123",
            storage_mode=CompressedDictStorageMode(),
        )
        result = entry.dict()
        assert result == {
            "fullUrl": "https://example.com/Patient/123",
            "resource": resource,
            "request": request.dict(),
            "response": response.dict(),
        }

    def test_from_dict_minimal(self) -> None:
        """Test creating from dictionary with minimal parameters."""
        data = {"resource": {"resourceType": "Patient"}}
        entry = FhirBundleEntry.from_dict(
            data, storage_mode=CompressedDictStorageMode()
        )

        assert entry.resource is not None
        with entry.resource.transaction():
            assert entry.resource.dict() == {"resourceType": "Patient"}
            assert entry.request is None
            assert entry.response is None
            assert entry.fullUrl is None

    def test_from_dict_full(self) -> None:
        """Test creating from dictionary with all parameters."""
        now = datetime.now(timezone.utc)
        data = {
            "fullUrl": "https://example.com/Patient/123",
            "resource": {"resourceType": "Patient", "id": "123"},
            "request": {"url": "https://example.com", "method": "GET"},
            "response": {
                "status": "200",
                "lastModified": now.isoformat(),
                "etag": 'W/"abc"',
            },
        }
        entry = FhirBundleEntry.from_dict(
            data, storage_mode=CompressedDictStorageMode()
        )
        assert entry.resource is not None
        with entry.resource.transaction():
            assert entry.fullUrl == "https://example.com/Patient/123"
            assert entry.resource.dict() == {"resourceType": "Patient", "id": "123"}
            assert entry.request is not None
            assert entry.request.url == "https://example.com"
            assert entry.response is not None
            assert entry.response.status == "200"

    def test_repr(self) -> None:
        """Test string representation of BundleEntry."""
        resource = {"resourceType": "Patient", "id": "123"}
        entry = FhirBundleEntry(
            resource=resource,
            request=None,
            response=None,
            storage_mode=CompressedDictStorageMode(),
        )
        assert repr(entry) == "BundleEntry(Patient/123)"



from datetime import datetime
from typing import Any, Dict, Optional, OrderedDict

from compressedfhir.utilities.json_helpers import FhirClientJsonHelpers


class FhirBundleEntryResponse:
    """
    FHIR Bundle Entry Response class for encapsulating the response from FHIR server when processing bundle entries
    """

    __slots__ = ["status", "lastModified", "etag", "location"]

    # noinspection PyPep8Naming
    def __init__(
        self,
        *,
        status: str = "200",
        etag: Optional[str] = None,
        lastModified: Optional[datetime] = None,
        location: Optional[str] = None,
    ) -> None:
        self.status: str = status or "200"
        if isinstance(status, int):
            self.status = str(status)
        self.lastModified: Optional[datetime] = lastModified
        self.etag: Optional[str] = etag
        self.location: Optional[str] = location

    def dict(self) -> OrderedDict[str, Any]:
        result: OrderedDict[str, Any] = OrderedDict[str, Any]({"status": self.status})
        if self.lastModified is not None:
            result["lastModified"] = self.lastModified.isoformat()
        if self.etag is not None:
            result["etag"] = self.etag
        if self.location is not None:
            result["location"] = self.location
        return FhirClientJsonHelpers.remove_empty_elements_from_ordered_dict(result)

    @classmethod
    def from_dict(
        cls, d: Dict[str, Any] | OrderedDict[str, Any]
    ) -> "FhirBundleEntryResponse":
        return cls(
            status=d["status"] if "status" in d else "200",
            lastModified=(
                datetime.fromisoformat(d["lastModified"])
                if "lastModified" in d
                else None
            ),
            etag=d["etag"] if "etag" in d else None,
            location=d["location"] if "location" in d else None,
        )

    def __deepcopy__(self, memo: Dict[int, Any]) -> "FhirBundleEntryResponse":
        return FhirBundleEntryResponse(
            status=self.status,
            lastModified=self.lastModified,
            etag=self.etag,
            location=self.location,
        )

    def __repr__(self) -> str:
        return f"FhirBundleEntryResponse(status: {self.status}, lastModified: {self.lastModified}, etag: {self.etag})"



import copy
import json
from typing import Any, Dict, Optional, List, OrderedDict

from compressedfhir.fhir.fhir_bundle_entry_search import FhirBundleEntrySearch
from compressedfhir.fhir.fhir_link import FhirLink
from compressedfhir.fhir.fhir_bundle_entry_request import FhirBundleEntryRequest
from compressedfhir.fhir.fhir_bundle_entry_response import (
    FhirBundleEntryResponse,
)
from compressedfhir.fhir.fhir_resource import FhirResource
from compressedfhir.utilities.compressed_dict.v1.compressed_dict import (
    CompressedDict,
)
from compressedfhir.utilities.compressed_dict.v1.compressed_dict_storage_mode import (
    CompressedDictStorageMode,
)
from compressedfhir.utilities.fhir_json_encoder import FhirJSONEncoder
from compressedfhir.utilities.json_helpers import FhirClientJsonHelpers


class FhirBundleEntry:
    """
    Represents a single entry in a FHIR Bundle.
    """

    __slots__ = [
        "_resource",
        "request",
        "response",
        "fullUrl",
        "link",
        "search",
        "storage_mode",
    ]

    # noinspection PyPep8Naming
    def __init__(
        self,
        *,
        fullUrl: Optional[str] = None,
        resource: Dict[str, Any] | FhirResource | OrderedDict[str, Any] | None,
        request: Optional[FhirBundleEntryRequest] = None,
        response: Optional[FhirBundleEntryResponse] = None,
        link: Optional[List[FhirLink]] = None,
        search: Optional[FhirBundleEntrySearch] = None,
        storage_mode: CompressedDictStorageMode = CompressedDictStorageMode.default(),
    ) -> None:
        """
        Initializes a BundleEntry object.

        :param fullUrl: The full URL of the entry.
        :param resource: The FHIR resource associated with the entry.
        :param request: The request information associated with the entry.
        :param response: The response information associated with the entry.
        :param storage_mode: The storage mode for the resource.
        """
        self._resource: Optional[FhirResource] = (
            resource
            if isinstance(resource, CompressedDict)
            else (
                FhirResource(initial_dict=resource, storage_mode=storage_mode)
                if resource is not None
                else None
            )
        )
        self.request: Optional[FhirBundleEntryRequest] = request
        self.response: Optional[FhirBundleEntryResponse] = response
        self.fullUrl: Optional[str] = fullUrl
        self.link: Optional[List[FhirLink]] = link
        self.search: Optional[FhirBundleEntrySearch] = search
        self.storage_mode: CompressedDictStorageMode = storage_mode

    # noinspection PyPep8Naming
    @classmethod
    def construct(
        cls,
        *,
        fullUrl: Optional[str] = None,
        resource: Dict[str, Any] | FhirResource | OrderedDict[str, Any] | None,
        request: Optional[FhirBundleEntryRequest] = None,
        response: Optional[FhirBundleEntryResponse] = None,
        link: Optional[List[FhirLink]] = None,
        search: Optional[FhirBundleEntrySearch] = None,
        storage_mode: CompressedDictStorageMode = CompressedDictStorageMode.default(),
    ) -> "FhirBundleEntry":
        """
        Constructs a BundleEntry object with the given parameters.

        :param fullUrl: The full URL of the entry.
        :param resource: The FHIR resource associated with the entry.
        :param request: The request information associated with the entry.
        :param response: The response information associated with the entry.
        :param link: The links associated with the entry.
        :param search: The search information associated with the entry.
        :param storage_mode: The storage mode for the resource.
        """
        return cls(
            fullUrl=fullUrl,
            resource=resource,
            request=request,
            response=response,
            link=link,
            search=search,
            storage_mode=storage_mode,
        )

    @property
    def resource(self) -> Optional[FhirResource]:
        """
        Returns the FHIR resource associated with the entry.

        :return: The FHIR resource.
        """
        return self._resource

    @resource.setter
    def resource(self, value: Dict[str, Any] | FhirResource | None) -> None:
        """
        Sets the FHIR resource associated with the entry.

        :param value: The FHIR resource to set.
        """
        if value is not None:
            if isinstance(value, CompressedDict):
                self._resource = value
            else:
                if self._resource is None:
                    self._resource = FhirResource(
                        initial_dict=value, storage_mode=self.storage_mode
                    )
                else:
                    self._resource.replace(value=value)
        else:
            self._resource = None

    def dict(self) -> OrderedDict[str, Any]:
        result: OrderedDict[str, Any] = OrderedDict[str, Any]()
        if self.fullUrl is not None:
            result["fullUrl"] = self.fullUrl
        if self.resource is not None:
            result["resource"] = self.resource.dict()
        if self.request is not None:
            result["request"] = self.request.dict()
        if self.response is not None:
            result["response"] = self.response.dict()
        if self.link is not None:
            result["link"] = [link.dict() for link in self.link]
        if self.search is not None:
            result["search"] = self.search.dict()
        return FhirClientJsonHelpers.remove_empty_elements_from_ordered_dict(result)

    @classmethod
    def from_dict(
        cls,
        d: OrderedDict[str, Any] | Dict[str, Any],
        storage_mode: CompressedDictStorageMode,
    ) -> "FhirBundleEntry":
        return cls(
            fullUrl=d["fullUrl"] if "fullUrl" in d else None,
            resource=(
                FhirResource(initial_dict=d["resource"], storage_mode=storage_mode)
                if "resource" in d
                else None
            ),
            request=(
                FhirBundleEntryRequest.from_dict(d["request"])
                if "request" in d
                else None
            ),
            response=(
                FhirBundleEntryResponse.from_dict(d["response"])
                if "response" in d
                else None
            ),
            link=(
                [FhirLink.from_dict(link) for link in d["link"]]
                if "link" in d
                else None
            ),
            search=(
                FhirBundleEntrySearch.from_dict(d["search"]) if "search" in d else None
            ),
            storage_mode=storage_mode,
        )

    def json(self) -> str:
        """
        Converts the BundleEntry object to a JSON string.

        :return: A JSON string representation of the BundleEntry.
        """
        return json.dumps(obj=self.dict(), cls=FhirJSONEncoder)

    def __deepcopy__(self, memo: Dict[int, Any]) -> "FhirBundleEntry":
        """
        Creates a copy of the BundleEntry object.

        :return: A new BundleEntry object with the same attributes.
        """
        return FhirBundleEntry(
            fullUrl=self.fullUrl,
            resource=copy.deepcopy(self.resource) if self.resource else None,
            request=copy.deepcopy(self.request) if self.request else None,
            response=copy.deepcopy(self.response) if self.response else None,
            link=[copy.deepcopy(link) for link in self.link] if self.link else None,
            search=copy.deepcopy(self.search) if self.search else None,
            storage_mode=self.storage_mode,
        )

    def __repr__(self) -> str:
        """
        Returns a string representation of the BundleEntry object.

        :return: A string representation of the BundleEntry.
        """
        return (
            f"BundleEntry({self.resource.resource_type}/{self.resource.id})"
            if self.resource
            else f"BundleEntry(Empty)"
        )

    def copy(self) -> "FhirBundleEntry":
        """
        Creates a copy of the BundleEntry object.

        :return: A new BundleEntry object with the same attributes.
        """
        return copy.deepcopy(self)

    @property
    def resource_type_and_id(self) -> Optional[str]:
        """Get the key from the resource."""
        return self.resource.resource_type_and_id if self.resource else None



import json
from typing import Dict, Any

from compressedfhir.fhir.fhir_resource import FhirResource
from compressedfhir.utilities.compressed_dict.v1.compressed_dict_storage_mode import (
    CompressedDictStorageMode,
)


class TestFhirResource:
    def test_init_empty(self) -> None:
        """Test initializing FhirResource with no initial dictionary."""
        resource = FhirResource(storage_mode=CompressedDictStorageMode())
        assert len(resource) == 0
        assert resource.resource_type is None
        assert resource.id is None
        assert resource.resource_type_and_id is None

    def test_init_with_data(self) -> None:
        """Test initializing FhirResource with a dictionary."""
        initial_data: Dict[str, Any] = {
            "resourceType": "Patient",
            "id": "123",
            "name": [{"given": ["John"]}],
        }
        resource = FhirResource(
            initial_dict=initial_data, storage_mode=CompressedDictStorageMode()
        )

        with resource.transaction():
            assert resource.resource_type == "Patient"
            assert resource.id == "123"
            assert resource.resource_type_and_id == "Patient/123"
            assert resource["name"][0]["given"][0] == "John"

    def test_resource_type_and_id_property(self) -> None:
        """Test resource_type_and_id property with various scenarios."""
        # Scenario 1: Both resource type and id present
        resource1 = FhirResource(
            initial_dict={"resourceType": "Observation", "id": "456"},
            storage_mode=CompressedDictStorageMode(),
        )
        assert resource1.resource_type_and_id == "Observation/456"

        # Scenario 2: Missing resource type
        resource2 = FhirResource(
            initial_dict={"id": "789"}, storage_mode=CompressedDictStorageMode()
        )
        assert resource2.resource_type_and_id is None

        # Scenario 3: Missing id
        resource3 = FhirResource(
            initial_dict={"resourceType": "Patient"},
            storage_mode=CompressedDictStorageMode(),
        )
        assert resource3.resource_type_and_id is None

    def test_equality(self) -> None:
        """Test equality comparison between FhirResource instances."""
        # Scenario 1: Equal resources
        resource1 = FhirResource(
            initial_dict={"resourceType": "Patient", "id": "123"},
            storage_mode=CompressedDictStorageMode(),
        )
        resource2 = FhirResource(
            initial_dict={"resourceType": "Patient", "id": "123"},
            storage_mode=CompressedDictStorageMode(),
        )
        assert resource1 == resource2

        # Scenario 2: Different resource types
        resource3 = FhirResource(
            initial_dict={"resourceType": "Observation", "id": "123"},
            storage_mode=CompressedDictStorageMode(),
        )
        assert resource1 != resource3

        # Scenario 3: Different ids
        resource4 = FhirResource(
            initial_dict={"resourceType": "Patient", "id": "456"},
            storage_mode=CompressedDictStorageMode(),
        )
        assert resource1 != resource4

        # Scenario 4: Comparing with non-FhirResource
        assert resource1 != "Not a FhirResource"

    def test_to_json(self) -> None:
        """Test JSON serialization of FhirResource."""
        initial_data: Dict[str, Any] = {
            "resourceType": "Patient",
            "id": "123",
            "name": [{"given": ["John"]}],
        }
        resource = FhirResource(
            initial_dict=initial_data, storage_mode=CompressedDictStorageMode()
        )

        json_str = resource.json()
        parsed_json = json.loads(json_str)

        assert parsed_json == initial_data
        assert "resourceType" in parsed_json
        assert "id" in parsed_json


class TestFhirResourceRemoveNulls:
    def test_remove_nulls_simple_dict(self) -> None:
        """
        Test removing None values from a simple dictionary
        """
        initial_dict: Dict[str, Any] = {
            "name": "John Doe",
            "age": None,
            "active": True,
            "email": None,
        }
        resource = FhirResource(initial_dict=initial_dict)
        resource.remove_nulls()

        with resource.transaction():
            # Check that None values are removed
            assert "age" not in resource
            assert "email" not in resource
            assert resource.get("name") == "John Doe"
            assert resource.get("active") is True

    def test_remove_nulls_nested_dict(self) -> None:
        """
        Test removing None values from a nested dictionary
        """
        initial_dict: Dict[str, Any] = {
            "patient": {
                "name": "Jane Smith",
                "contact": None,
                "address": {"street": None, "city": "New York"},
            },
            "status": None,
        }
        resource = FhirResource(initial_dict=initial_dict)
        resource.remove_nulls()

        with resource.transaction():
            assert "status" not in resource
            assert "contact" not in resource.get("patient", {})
            assert resource.get("patient", {}).get("address", {}).get("street") is None
            assert (
                resource.get("patient", {}).get("address", {}).get("city") == "New York"
            )

    def test_remove_nulls_list_of_dicts(self) -> None:
        """
        Test removing None values from a list of dictionaries
        """
        initial_dict: Dict[str, Any] = {
            "patients": [
                {"name": "Alice", "age": None},
                {"name": "Bob", "age": 30},
                {"name": None, "active": False},
            ]
        }
        resource = FhirResource(initial_dict=initial_dict)
        resource.remove_nulls()

        with resource.transaction():
            assert len(resource.get("patients", [])) == 3
            assert resource.get("patients", [])[0].get("name") == "Alice"
            assert resource.get("patients", [])[1].get("name") == "Bob"
            assert resource.get("patients", [])[1].get("age") == 30

    def test_remove_nulls_empty_dict(self) -> None:
        """
        Test removing None values from an empty dictionary
        """
        resource = FhirResource(initial_dict={})
        resource.remove_nulls()

        assert len(resource) == 0

    def test_remove_nulls_no_changes(self) -> None:
        """
        Test removing None values when no None values exist
        """
        initial_dict: Dict[str, Any] = {
            "name": "Test User",
            "active": True,
            "score": 100,
        }
        resource = FhirResource(initial_dict=initial_dict)
        original_dict = resource.copy()
        resource.remove_nulls()

        assert resource == original_dict

    def test_remove_nulls_with_custom_storage_mode(self) -> None:
        """
        Test removing None values with a custom storage mode
        """
        initial_dict: Dict[str, Any] = {
            "name": "Custom Mode User",
            "email": None,
            "active": True,
        }
        resource = FhirResource(
            initial_dict=initial_dict, storage_mode=CompressedDictStorageMode.default()
        )
        resource.remove_nulls()

        with resource.transaction():
            assert "email" not in resource
            assert resource.get("name") == "Custom Mode User"
            assert resource.get("active") is True

    def test_remove_nulls_preserves_false_and_zero_values(self) -> None:
        """
        Test that False and 0 values are not removed
        """
        initial_dict: Dict[str, Any] = {"active": False, "score": 0, "name": None}
        resource = FhirResource(initial_dict=initial_dict)
        resource.remove_nulls()

        with resource.transaction():
            assert resource.get("active") is False
            assert resource.get("score") == 0
            assert "name" not in resource



import pytest
from compressedfhir.fhir.fhir_bundle_entry import FhirBundleEntry
from compressedfhir.fhir.fhir_bundle_entry_list import FhirBundleEntryList
from compressedfhir.utilities.compressed_dict.v1.compressed_dict_storage_mode import (
    CompressedDictStorageMode,
)


@pytest.mark.asyncio
class TestFhirBundleEntryList:
    async def test_consume_resource_async_with_none_batch_size(self) -> None:
        """
        Test consuming bundle entries when batch_size is None.
        Each call should yield a single entry.
        """
        # Arrange
        entries = [
            FhirBundleEntry(
                resource={"id": "1"},
                request=None,
                response=None,
                fullUrl=None,
                storage_mode=CompressedDictStorageMode(),
            ),
            FhirBundleEntry(
                resource={"id": "2"},
                request=None,
                response=None,
                fullUrl=None,
                storage_mode=CompressedDictStorageMode(),
            ),
            FhirBundleEntry(
                resource={"id": "3"},
                request=None,
                response=None,
                fullUrl=None,
                storage_mode=CompressedDictStorageMode(),
            ),
        ]
        bundle_list = FhirBundleEntryList(entries)

        # Act
        batches = []
        async for batch in bundle_list.consume_resource_async(batch_size=None):
            batches.append(batch)

        # Assert
        assert len(batches) == 3
        assert all(len(batch) == 1 for batch in batches)
        assert len(bundle_list) == 0  # All entries should be consumed

    @pytest.mark.parametrize("batch_size", [1, 2, 3, 5])
    async def test_consume_resource_async_with_valid_batch_sizes(
        self, batch_size: int
    ) -> None:
        """
        Test consuming bundle entries with various valid batch sizes.
        """
        # Arrange
        entries = [
            FhirBundleEntry(
                resource={"id": "1"},
                request=None,
                response=None,
                fullUrl=None,
                storage_mode=CompressedDictStorageMode(),
            ),
            FhirBundleEntry(
                resource={"id": "2"},
                request=None,
                response=None,
                fullUrl=None,
                storage_mode=CompressedDictStorageMode(),
            ),
            FhirBundleEntry(
                resource={"id": "3"},
                request=None,
                response=None,
                fullUrl=None,
                storage_mode=CompressedDictStorageMode(),
            ),
            FhirBundleEntry(
                resource={"id": "4"},
                request=None,
                response=None,
                fullUrl=None,
                storage_mode=CompressedDictStorageMode(),
            ),
            FhirBundleEntry(
                resource={"id": "5"},
                request=None,
                response=None,
                fullUrl=None,
                storage_mode=CompressedDictStorageMode(),
            ),
        ]
        bundle_list = FhirBundleEntryList(entries)

        # Act
        batches = []
        async for batch in bundle_list.consume_resource_async(batch_size=batch_size):
            batches.append(batch)

        # Assert
        expected_batch_count = (len(entries) + batch_size - 1) // batch_size
        assert len(batches) == expected_batch_count
        assert all(len(batch) <= batch_size for batch in batches)
        assert sum(len(batch) for batch in batches) == len(entries)
        assert len(bundle_list) == 0  # All entries should be consumed

    async def test_consume_resource_async_with_zero_batch_size(self) -> None:
        """
        Test that consuming with batch_size <= 0 raises a ValueError.
        """
        # Arrange
        entries = [
            FhirBundleEntry(
                resource={"id": "1"},
                request=None,
                response=None,
                fullUrl=None,
                storage_mode=CompressedDictStorageMode(),
            )
        ]
        bundle_list = FhirBundleEntryList(entries)

        # Act & Assert
        with pytest.raises(ValueError, match="Batch size must be greater than 0."):
            async for _ in bundle_list.consume_resource_async(batch_size=0):
                pass

    async def test_consume_resource_async_with_empty_list(self) -> None:
        """
        Test consuming an empty list.
        """
        # Arrange
        bundle_list = FhirBundleEntryList()

        # Act
        batches = []
        async for batch in bundle_list.consume_resource_async(batch_size=2):
            batches.append(batch)

        # Assert
        assert len(batches) == 0

    def test_class_inheritance(self) -> None:
        """
        Verify that FhirBundleEntryList inherits from collections.deque
        and supports deque operations.
        """
        # Arrange
        entries = [
            FhirBundleEntry(
                resource={"id": "1"},
                request=None,
                response=None,
                fullUrl=None,
                storage_mode=CompressedDictStorageMode(),
            ),
            FhirBundleEntry(
                resource={"id": "2"},
                request=None,
                response=None,
                fullUrl=None,
                storage_mode=CompressedDictStorageMode(),
            ),
        ]

        # Act
        bundle_list = FhirBundleEntryList(entries)

        # Assert
        assert len(bundle_list) == 2
        bundle_list.append(
            FhirBundleEntry(
                resource={"id": "3"},
                request=None,
                response=None,
                fullUrl=None,
                storage_mode=CompressedDictStorageMode(),
            )
        )
        assert len(bundle_list) == 3
        resource = bundle_list.popleft().resource
        assert resource is not None
        assert resource.dict() == {"id": "1"}
